### Program for the Ring Stacker robot to solve the Towers of Hanoi puzzle with 4 rings

from mindstorms import MSHub, Motor, MotorPair, ColorSensor, DistanceSensor, App
from mindstorms.control import wait_for_seconds, wait_until, Timer
from mindstorms.operator import greater_than, greater_than_or_equal_to, less_than, less_than_or_equal_to, equal_to, not_equal_to
import math


# Variables used by the program as a whole.
hub = MSHub()
claw = Motor('B')
lift = Motor('F')
turntable = Motor('D')
ring_counts = [0, 4, 0, 0]     # initial ring counts indexed by pole (1-3) (index 0 is not used)
LEVEL_ANGLES = [0, 335, 348, 355, 9]    # lift motor position to align with indexed stack height (index 0 is not used)
current_pole = 2        # arm starts centered on center pole (left pole is 1, center is 2, right is 3) 


# Open the claw
def open_claw():
    claw.run_to_position(15)


# Close the claw
def close_claw():
    claw.run_to_position(350)


# Raise the arm up above the top of the poles
def raise_arm():
    lift.run_to_position(100)


# Lower the arm to align with a ring at the specified level (1-4)
def lower_arm(level):
    lift.run_to_position(LEVEL_ANGLES[level])


# Rotate the arm to align with the center of the specified pole (1-3), starting from current_pole (1-3)
def rotate_to_pole(pole):
    global current_pole
    turntable.run_for_degrees((current_pole - pole) * 270)  # 90 degrees per pole with 3:1 turntable gear ratio
    current_pole = pole


# Update the display to add (brightness = 0) or remove (brightness = 0) a pixel
# from the specified pole at the specified height level.
def change_pixel(pole, level, brightness):
    hub.light_matrix.set_pixel(level - 1, (pole - 1) * 2, brightness)


# Pick up a ring from the specified pole (1-3).
def get_from_pole(pole):
    rotate_to_pole(pole)
    count = ring_counts[pole]
    lower_arm(count)
    close_claw()
    raise_arm()
    change_pixel(pole, count, 0)
    ring_counts[pole] = count - 1


# Place a held ring on the specified pole (1-3).
def put_on_pole(pole):
    rotate_to_pole(pole)
    count = ring_counts[pole] + 1
    lower_arm(count)
    change_pixel(pole, count, 100)
    open_claw()
    raise_arm()
    ring_counts[pole] = count


# Move a ring from pole src (1-3) to pole dest (1-3)
def move_ring(src, dest):
    get_from_pole(src)
    put_on_pole(dest)


# Move a stack of count rings from pole src (1-3) to pole dest (1-3)
# with pole other (1-3) as the temporary pole.
# Using the classic recursive solution to the Towers of Hanoi puzzle,
# this involves a recursive move of count - 1 rings from src to other,
# then moving a single ring from src to dest, then another recursive 
# move of count - 1 rings from other to dest.
def move_stack(count, src, dest, other):
    if count > 0:
        move_stack(count - 1, src, other, dest)
        move_ring(src, dest)
        move_stack(count - 1, other, dest, src)


# The main program starts here
hub.light_matrix.off()
hub.light_matrix.set_pixel(0, 0, 100)   # init display of 4 pixel stack on the left side
hub.light_matrix.set_pixel(1, 0, 100)
hub.light_matrix.set_pixel(2, 0, 100)
hub.light_matrix.set_pixel(3, 0, 100)
lift.set_default_speed(40)
raise_arm()
open_claw()
move_stack(4, 1, 3, 2)    # move 4 rings from pole 1 to pole 3 using pole 2 as the other pole
rotate_to_pole(2)         # put arm back where it started
